# -*- coding: utf-8 -*-

"""Top-level package for LyricsMaster."""

__author__ = """SekouD"""
__email__ = 'sekoud.python@gmail.com'
__version__ = '2.7.21'

from .providers import LyricWiki, AzLyrics, Genius
from .utils import TorController
